@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 from math import sqrt
 
 import numpy as np
@@ -21,6 +19,8 @@ class AverageLearner(BaseLearner):
         Desired absolute tolerance.
     rtol : float
         Desired relative tolerance.
+    min_npoints : int
+        Minimum number of points to sample.
 
     Attributes
     ----------
@@ -32,9 +32,9 @@ class AverageLearner(BaseLearner):
         Number of evaluated points.
     """
 
-    def __init__(self, function, atol=None, rtol=None):
+    def __init__(self, function, atol=None, rtol=None, min_npoints=2):
         if atol is None and rtol is None:
-            raise Exception('At least one of `atol` and `rtol` should be set.')
+            raise Exception("At least one of `atol` and `rtol` should be set.")
         if atol is None:
             atol = np.inf
         if rtol is None:
@@ -46,6 +46,8 @@ class AverageLearner(BaseLearner):
         self.atol = atol
         self.rtol = rtol
         self.npoints = 0
+        # Cannot estimate standard deviation with fewer than 2 points.
+        self.min_npoints = max(min_npoints, 2)
         self.sum_f = 0
         self.sum_f_sq = 0
 
@@ -58,9 +60,11 @@ class AverageLearner(BaseLearner):
 
         if any(p in self.data or p in self.pending_points for p in points):
             # This means some of the points `< self.n_requested` do not exist.
-            points = list(set(range(self.n_requested + n))
-                          - set(self.data)
-                          - set(self.pending_points))[:n]
+            points = list(
+                set(range(self.n_requested + n))
+                - set(self.data)
+                - set(self.pending_points)
+            )[:n]
 
         loss_improvements = [self._loss_improvement(n) / n] * n
         if tell_pending:
@@ -76,7 +80,7 @@ class AverageLearner(BaseLearner):
         self.data[n] = value
         self.pending_points.discard(n)
         self.sum_f += value
-        self.sum_f_sq += value**2
+        self.sum_f_sq += value ** 2
         self.npoints += 1
 
     def tell_pending(self, n):
@@ -92,9 +96,9 @@ class AverageLearner(BaseLearner):
         """The corrected sample standard deviation of the values
         in `data`."""
         n = self.npoints
-        if n < 2:
+        if n < self.min_npoints:
             return np.inf
-        numerator = self.sum_f_sq - n * self.mean**2
+        numerator = self.sum_f_sq - n * self.mean ** 2
         if numerator < 0:
             # in this case the numerator ~ -1e-15
             return 0
@@ -106,11 +110,12 @@ class AverageLearner(BaseLearner):
             n = self.npoints if real else self.n_requested
         else:
             n = n
-        if n < 2:
+        if n < self.min_npoints:
             return np.inf
         standard_error = self.std / sqrt(n)
-        return max(standard_error / self.atol,
-                   standard_error / abs(self.mean) / self.rtol)
+        return max(
+            standard_error / self.atol, standard_error / abs(self.mean) / self.rtol
+        )
 
     def _loss_improvement(self, n):
         loss = self.loss()
@@ -136,10 +141,24 @@ class AverageLearner(BaseLearner):
             return hv.Histogram([[], []])
         num_bins = int(max(5, sqrt(self.npoints)))
         vals = hv.Points(vals)
-        return hv.operation.histogram(vals, num_bins=num_bins, dimension=1)
+        return hv.operation.histogram(vals, num_bins=num_bins, dimension="y")
 
     def _get_data(self):
         return (self.data, self.npoints, self.sum_f, self.sum_f_sq)
 
     def _set_data(self, data):
         self.data, self.npoints, self.sum_f, self.sum_f_sq = data
+
+    def __getstate__(self):
+        return (
+            self.function,
+            self.atol,
+            self.rtol,
+            self.min_npoints,
+            self._get_data(),
+        )
+
+    def __setstate__(self, state):
+        function, atol, rtol, min_npoints, data = state
+        self.__init__(function, atol, rtol, min_npoints)
+        self._set_data(data)
