@@ -1,61 +1,58 @@
-# -*- coding: utf-8 -*-
-
 import abc
 import asyncio
 import concurrent.futures as concurrent
-from contextlib import suppress
 import functools
 import inspect
-import os
+import itertools
+import pickle
 import time
 import traceback
 import warnings
+from contextlib import suppress
 
-from adaptive.notebook_integration import live_plot, live_info, in_ipynb
+from adaptive.notebook_integration import in_ipynb, live_info, live_plot
 
 try:
     import ipyparallel
+
     with_ipyparallel = True
 except ModuleNotFoundError:
     with_ipyparallel = False
 
 try:
     import distributed
+
     with_distributed = True
 except ModuleNotFoundError:
     with_distributed = False
 
+try:
+    import mpi4py.futures
+
+    with_mpi4py = True
+except ModuleNotFoundError:
+    with_mpi4py = False
+
+try:
+    import loky
+
+    with_loky = True
+except ModuleNotFoundError:
+    with_loky = False
+
 with suppress(ModuleNotFoundError):
     import uvloop
+
     asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
 
 
-if os.name == 'nt':
-    if with_distributed:
-        _default_executor = distributed.Client
-        _default_executor_kwargs = {'address': distributed.LocalCluster()}
-    else:
-        _windows_executor_msg = (
-            "The default executor on Windows for 'adaptive.Runner' cannot "
-            "be used because the package 'distributed' is not installed. "
-            "Either install 'distributed' or explicitly specify an executor "
-            "when using 'adaptive.Runner'."
-        )
-
-        _default_executor_kwargs = {}
-
-        def _default_executor(*args, **kwargs):
-            raise RuntimeError(_windows_executor_msg)
-
-        warnings.warn(_windows_executor_msg)
-
-else:
-    _default_executor = concurrent.ProcessPoolExecutor
-    _default_executor_kwargs = {}
+_default_executor = (
+    loky.get_reusable_executor if with_loky else concurrent.ProcessPoolExecutor
+)
 
 
 class BaseRunner(metaclass=abc.ABCMeta):
-    """Base class for runners that use `concurrent.futures.Executors`.
+    r"""Base class for runners that use `concurrent.futures.Executors`.
 
     Parameters
     ----------
@@ -65,11 +62,9 @@ class BaseRunner(metaclass=abc.ABCMeta):
         the learner as its sole argument, and return True when we should
         stop requesting more points.
     executor : `concurrent.futures.Executor`, `distributed.Client`,\
-               or `ipyparallel.Client`, optional
+               `mpi4py.futures.MPIPoolExecutor`, or `ipyparallel.Client`, optional
         The executor in which to evaluate the function to be learned.
-        If not provided, a new `~concurrent.futures.ProcessPoolExecutor`
-        is used on Unix systems while on Windows a `distributed.Client`
-        is used if `distributed` is installed.
+        If not provided, a new `~concurrent.futures.ProcessPoolExecutor`.
     ntasks : int, optional
         The number of concurrent function evaluations. Defaults to the number
         of cores available in `executor`.
@@ -93,14 +88,14 @@ class BaseRunner(metaclass=abc.ABCMeta):
     log : list or None
         Record of the method calls made to the learner, in the format
         ``(method_name, *args)``.
-    to_retry : dict
-        Mapping of ``{point: n_fails, ...}``. When a point has failed
+    to_retry : list of tuples
+        List of ``(point, n_fails)``. When a point has failed
         ``runner.retries`` times it is removed but will be present
         in ``runner.tracebacks``.
-    tracebacks : dict
-        A mapping of point to the traceback if that point failed.
-    pending_points : dict
-        A mapping of `~concurrent.futures.Future`\s to points.
+    tracebacks : list of tuples
+        List of of ``(point, tb)`` for points that failed.
+    pending_points : list of tuples
+        A list of tuples with ``(concurrent.futures.Future, point)``.
 
     Methods
     -------
@@ -110,17 +105,25 @@ class BaseRunner(metaclass=abc.ABCMeta):
 
     """
 
-    def __init__(self, learner, goal, *,
-                 executor=None, ntasks=None, log=False,
-                 shutdown_executor=False, retries=0,
-                 raise_if_retries_exceeded=True):
+    def __init__(
+        self,
+        learner,
+        goal,
+        *,
+        executor=None,
+        ntasks=None,
+        log=False,
+        shutdown_executor=False,
+        retries=0,
+        raise_if_retries_exceeded=True,
+    ):
 
         self.executor = _ensure_executor(executor)
         self.goal = goal
 
         self._max_tasks = ntasks
 
-        self.pending_points = {}
+        self._pending_tasks = {}  # mapping from concurrent.futures.Future → point id
 
         # if we instantiate our own executor, then we are also responsible
         # for calling 'shutdown'
@@ -137,18 +140,24 @@ class BaseRunner(metaclass=abc.ABCMeta):
         # Error handling attributes
         self.retries = retries
         self.raise_if_retries_exceeded = raise_if_retries_exceeded
-        self.to_retry = {}
-        self.tracebacks = {}
+        self._to_retry = {}
+        self._tracebacks = {}
+
+        self._id_to_point = {}
+        self._next_id = functools.partial(
+            next, itertools.count()
+        )  # some unique id to be associated with each point
 
     def _get_max_tasks(self):
         return self._max_tasks or _get_ncores(self.executor)
 
-    def _do_raise(self, e, x):
-        tb = self.tracebacks[x]
+    def _do_raise(self, e, i):
+        tb = self._tracebacks[i]
+        x = self._id_to_point[i]
         raise RuntimeError(
-            'An error occured while evaluating '
+            "An error occured while evaluating "
             f'"learner.function({x})". '
-            f'See the traceback for details.:\n\n{tb}'
+            f"See the traceback for details.:\n\n{tb}"
         ) from e
 
     @property
@@ -156,14 +165,21 @@ class BaseRunner(metaclass=abc.ABCMeta):
         return self.log is not None
 
     def _ask(self, n):
-        points = [p for p in self.to_retry.keys()
-                  if p not in self.pending_points.values()][:n]
-        loss_improvements = len(points) * [float('inf')]
-        if len(points) < n:
-            p, l = self.learner.ask(n - len(points))
-            points += p
-            loss_improvements += l
-        return points, loss_improvements
+        pending_ids = self._pending_tasks.values()
+        # using generator here because we only need until `n`
+        pids_gen = (pid for pid in self._to_retry.keys() if pid not in pending_ids)
+        pids = list(itertools.islice(pids_gen, n))
+
+        loss_improvements = len(pids) * [float("inf")]
+
+        if len(pids) < n:
+            new_points, new_losses = self.learner.ask(n - len(pids))
+            loss_improvements += new_losses
+            for point in new_points:
+                pid = self._next_id()
+                self._id_to_point[pid] = point
+                pids.append(pid)
+        return pids, loss_improvements
 
     def overhead(self):
         """Overhead of using Adaptive and the executor in percent.
@@ -190,79 +206,101 @@ class BaseRunner(metaclass=abc.ABCMeta):
 
     def _process_futures(self, done_futs):
         for fut in done_futs:
-            x = self.pending_points.pop(fut)
+            pid = self._pending_tasks.pop(fut)
             try:
                 y = fut.result()
                 t = time.time() - fut.start_time  # total execution time
             except Exception as e:
-                self.tracebacks[x] = traceback.format_exc()
-                self.to_retry[x] = self.to_retry.get(x, 0) + 1
-                if self.to_retry[x] > self.retries:
-                    self.to_retry.pop(x)
+                self._tracebacks[pid] = traceback.format_exc()
+                self._to_retry[pid] = self._to_retry.get(pid, 0) + 1
+                if self._to_retry[pid] > self.retries:
+                    self._to_retry.pop(pid)
                     if self.raise_if_retries_exceeded:
-                        self._do_raise(e, x)
+                        self._do_raise(e, pid)
             else:
                 self._elapsed_function_time += t / self._get_max_tasks()
-                self.to_retry.pop(x, None)
-                self.tracebacks.pop(x, None)
+                self._to_retry.pop(pid, None)
+                self._tracebacks.pop(pid, None)
+                x = self._id_to_point.pop(pid)
                 if self.do_log:
-                    self.log.append(('tell', x, y))
+                    self.log.append(("tell", x, y))
                 self.learner.tell(x, y)
 
     def _get_futures(self):
         # Launch tasks to replace the ones that completed
         # on the last iteration, making sure to fill workers
         # that have started since the last iteration.
-        n_new_tasks = max(0, self._get_max_tasks() - len(self.pending_points))
+        n_new_tasks = max(0, self._get_max_tasks() - len(self._pending_tasks))
 
         if self.do_log:
-            self.log.append(('ask', n_new_tasks))
+            self.log.append(("ask", n_new_tasks))
 
-        points, _ = self._ask(n_new_tasks)
+        pids, _ = self._ask(n_new_tasks)
 
-        for x in points:
+        for pid in pids:
             start_time = time.time()  # so we can measure execution time
-            fut = self._submit(x)
+            point = self._id_to_point[pid]
+            fut = self._submit(point)
             fut.start_time = start_time
-            self.pending_points[fut] = x
+            self._pending_tasks[fut] = pid
 
         # Collect and results and add them to the learner
-        futures = list(self.pending_points.keys())
+        futures = list(self._pending_tasks.keys())
         return futures
 
     def _remove_unfinished(self):
         # remove points with 'None' values from the learner
         self.learner.remove_unfinished()
         # cancel any outstanding tasks
-        remaining = list(self.pending_points.keys())
+        remaining = list(self._pending_tasks.keys())
         for fut in remaining:
             fut.cancel()
         return remaining
 
     def _cleanup(self):
         if self.shutdown_executor:
-            self.executor.shutdown(wait=False)
+            # XXX: temporary set wait=True because of a bug with Python ≥3.7
+            # and loky in any Python version.
+            # see https://github.com/python-adaptive/adaptive/issues/156
+            # and https://github.com/python-adaptive/adaptive/pull/164
+            # and https://bugs.python.org/issue36281
+            # and https://github.com/joblib/loky/issues/241
+            self.executor.shutdown(wait=True)
         self.end_time = time.time()
 
     @property
     def failed(self):
         """Set of points that failed ``runner.retries`` times."""
-        return set(self.tracebacks) - set(self.to_retry)
-    
+        return set(self._tracebacks) - set(self._to_retry)
+
     @abc.abstractmethod
     def elapsed_time(self):
         """Return the total time elapsed since the runner
         was started.
-        
+
         Is called in `overhead`.
         """
         pass
-    
+
     @abc.abstractmethod
     def _submit(self, x):
         """Is called in `_get_futures`."""
         pass
 
+    @property
+    def tracebacks(self):
+        return [(self._id_to_point[pid], tb) for pid, tb in self._tracebacks.items()]
+
+    @property
+    def to_retry(self):
+        return [(self._id_to_point[pid], n) for pid, n in self._to_retry.items()]
+
+    @property
+    def pending_points(self):
+        return [
+            (fut, self._id_to_point[pid]) for fut, pid in self._pending_tasks.items()
+        ]
+
 
 class BlockingRunner(BaseRunner):
     """Run a learner synchronously in an executor.
@@ -275,11 +313,10 @@ class BlockingRunner(BaseRunner):
         the learner as its sole argument, and return True when we should
         stop requesting more points.
     executor : `concurrent.futures.Executor`, `distributed.Client`,\
-               or `ipyparallel.Client`, optional
+               `mpi4py.futures.MPIPoolExecutor`, `ipyparallel.Client` or\
+               `loky.get_reusable_executor`, optional
         The executor in which to evaluate the function to be learned.
-        If not provided, a new `~concurrent.futures.ProcessPoolExecutor`
-        is used on Unix systems while on Windows a `distributed.Client`
-        is used if `distributed` is installed.
+        If not provided, a new `~concurrent.futures.ProcessPoolExecutor`.
     ntasks : int, optional
         The number of concurrent function evaluations. Defaults to the number
         of cores available in `executor`.
@@ -303,14 +340,14 @@ class BlockingRunner(BaseRunner):
     log : list or None
         Record of the method calls made to the learner, in the format
         ``(method_name, *args)``.
-    to_retry : dict
-        Mapping of ``{point: n_fails, ...}``. When a point has failed
+    to_retry : list of tuples
+        List of ``(point, n_fails)``. When a point has failed
         ``runner.retries`` times it is removed but will be present
         in ``runner.tracebacks``.
-    tracebacks : dict
-        A mapping of point to the traceback if that point failed.
-    pending_points : dict
-        A mapping of `~concurrent.futures.Future`\to points.
+    tracebacks : list of tuples
+        List of of ``(point, tb)`` for points that failed.
+    pending_points : list of tuples
+        A list of tuples with ``(concurrent.futures.Future, point)``.
 
     Methods
     -------
@@ -324,17 +361,30 @@ class BlockingRunner(BaseRunner):
 
     """
 
-    def __init__(self, learner, goal, *,
-                 executor=None, ntasks=None, log=False,
-                 shutdown_executor=False, retries=0,
-                 raise_if_retries_exceeded=True):
+    def __init__(
+        self,
+        learner,
+        goal,
+        *,
+        executor=None,
+        ntasks=None,
+        log=False,
+        shutdown_executor=False,
+        retries=0,
+        raise_if_retries_exceeded=True,
+    ):
         if inspect.iscoroutinefunction(learner.function):
-            raise ValueError("Coroutine functions can only be used "
-                             "with 'AsyncRunner'.")
-        super().__init__(learner, goal, executor=executor, ntasks=ntasks,
-                         log=log, shutdown_executor=shutdown_executor,
-                         retries=retries,
-                         raise_if_retries_exceeded=raise_if_retries_exceeded)
+            raise ValueError("Coroutine functions can only be used with 'AsyncRunner'.")
+        super().__init__(
+            learner,
+            goal,
+            executor=executor,
+            ntasks=ntasks,
+            log=log,
+            shutdown_executor=shutdown_executor,
+            retries=retries,
+            raise_if_retries_exceeded=raise_if_retries_exceeded,
+        )
         self._run()
 
     def _submit(self, x):
@@ -344,13 +394,12 @@ class BlockingRunner(BaseRunner):
         first_completed = concurrent.FIRST_COMPLETED
 
         if self._get_max_tasks() < 1:
-            raise RuntimeError('Executor has no workers')
+            raise RuntimeError("Executor has no workers")
 
         try:
             while not self.goal(self.learner):
                 futures = self._get_futures()
-                done, _ = concurrent.wait(futures,
-                                          return_when=first_completed)
+                done, _ = concurrent.wait(futures, return_when=first_completed)
                 self._process_futures(done)
         finally:
             remaining = self._remove_unfinished()
@@ -369,7 +418,7 @@ class BlockingRunner(BaseRunner):
 
 
 class AsyncRunner(BaseRunner):
-    """Run a learner asynchronously in an executor using `asyncio`.
+    r"""Run a learner asynchronously in an executor using `asyncio`.
 
     Parameters
     ----------
@@ -380,11 +429,10 @@ class AsyncRunner(BaseRunner):
         stop requesting more points. If not provided, the runner will run
         forever, or until ``self.task.cancel()`` is called.
     executor : `concurrent.futures.Executor`, `distributed.Client`,\
-               or `ipyparallel.Client`, optional
+               `mpi4py.futures.MPIPoolExecutor`, `ipyparallel.Client` or\
+               `loky.get_reusable_executor`, optional
         The executor in which to evaluate the function to be learned.
-        If not provided, a new `~concurrent.futures.ProcessPoolExecutor`
-        is used on Unix systems while on Windows a `distributed.Client`
-        is used if `distributed` is installed.
+        If not provided, a new `~concurrent.futures.ProcessPoolExecutor`.
     ntasks : int, optional
         The number of concurrent function evaluations. Defaults to the number
         of cores available in `executor`.
@@ -413,14 +461,14 @@ class AsyncRunner(BaseRunner):
     log : list or None
         Record of the method calls made to the learner, in the format
         ``(method_name, *args)``.
-    to_retry : dict
-        Mapping of ``{point: n_fails, ...}``. When a point has failed
+    to_retry : list of tuples
+        List of ``(point, n_fails)``. When a point has failed
         ``runner.retries`` times it is removed but will be present
         in ``runner.tracebacks``.
-    tracebacks : dict
-        A mapping of point to the traceback if that point failed.
-    pending_points : dict
-        A mapping of `~concurrent.futures.Future`\s to points.
+    tracebacks : list of tuples
+        List of of ``(point, tb)`` for points that failed.
+    pending_points : list of tuples
+        A list of tuples with ``(concurrent.futures.Future, point)``.
 
     Methods
     -------
@@ -439,19 +487,51 @@ class AsyncRunner(BaseRunner):
     run directly on the event loop (and not in the executor).
     """
 
-    def __init__(self, learner, goal=None, *,
-                 executor=None, ntasks=None, log=False,
-                 shutdown_executor=False, ioloop=None,
-                 retries=0, raise_if_retries_exceeded=True):
+    def __init__(
+        self,
+        learner,
+        goal=None,
+        *,
+        executor=None,
+        ntasks=None,
+        log=False,
+        shutdown_executor=False,
+        ioloop=None,
+        retries=0,
+        raise_if_retries_exceeded=True,
+    ):
 
         if goal is None:
+
             def goal(_):
                 return False
 
-        super().__init__(learner, goal, executor=executor, ntasks=ntasks,
-                         log=log, shutdown_executor=shutdown_executor,
-                         retries=retries,
-                         raise_if_retries_exceeded=raise_if_retries_exceeded)
+        if (
+            executor is None
+            and _default_executor is concurrent.ProcessPoolExecutor
+            and not inspect.iscoroutinefunction(learner.function)
+        ):
+            try:
+                pickle.dumps(learner.function)
+            except pickle.PicklingError:
+                raise ValueError(
+                    "`learner.function` cannot be pickled (is it a lamdba function?)"
+                    " and therefore does not work with the default executor."
+                    " Either make sure the function is pickleble or use an executor"
+                    " that might work with 'hard to pickle'-functions"
+                    " , e.g. `ipyparallel` with `dill`."
+                )
+
+        super().__init__(
+            learner,
+            goal,
+            executor=executor,
+            ntasks=ntasks,
+            log=log,
+            shutdown_executor=shutdown_executor,
+            retries=retries,
+            raise_if_retries_exceeded=raise_if_retries_exceeded,
+        )
         self.ioloop = ioloop or asyncio.get_event_loop()
         self.task = None
 
@@ -461,17 +541,20 @@ class AsyncRunner(BaseRunner):
         # the user can have more fine-grained control over the parallelism.
         if inspect.iscoroutinefunction(learner.function):
             if executor:  # user-provided argument
-                raise RuntimeError('Cannot use an executor when learning an '
-                                   'async function.')
+                raise RuntimeError(
+                    "Cannot use an executor when learning an async function."
+                )
             self.executor.shutdown()  # Make sure we don't shoot ourselves later
 
         self.task = self.ioloop.create_task(self._run())
         self.saving_task = None
         if in_ipynb() and not self.ioloop.is_running():
-            warnings.warn("The runner has been scheduled, but the asyncio "
-                          "event loop is not running! If you are "
-                          "in a Jupyter notebook, remember to run "
-                          "'adaptive.notebook_extension()'")
+            warnings.warn(
+                "The runner has been scheduled, but the asyncio "
+                "event loop is not running! If you are "
+                "in a Jupyter notebook, remember to run "
+                "'adaptive.notebook_extension()'"
+            )
 
     def _submit(self, x):
         ioloop = self.ioloop
@@ -488,13 +571,13 @@ class AsyncRunner(BaseRunner):
         try:
             self.task.result()
         except asyncio.CancelledError:
-            return 'cancelled'
+            return "cancelled"
         except asyncio.InvalidStateError:
-            return 'running'
+            return "running"
         except Exception:
-            return 'failed'
+            return "failed"
         else:
-            return 'finished'
+            return "finished"
 
     def cancel(self):
         """Cancel the runner.
@@ -503,12 +586,12 @@ class AsyncRunner(BaseRunner):
         """
         self.task.cancel()
 
-    def live_plot(self, *, plotter=None, update_interval=2, name=None):
+    def live_plot(self, *, plotter=None, update_interval=2, name=None, normalize=True):
         """Live plotting of the learner's data.
 
         Parameters
         ----------
-        runner : `Runner`
+        runner : `~adaptive.Runner`
         plotter : function
             A function that takes the learner as a argument and returns a
             holoviews object. By default ``learner.plot()`` will be called.
@@ -518,15 +601,17 @@ class AsyncRunner(BaseRunner):
             Name for the `live_plot` task in `adaptive.active_plotting_tasks`.
             By default the name is None and if another task with the same name
             already exists that other `live_plot` is canceled.
+        normalize : bool
+            Normalize (scale to fit) the frame upon each update.
 
         Returns
         -------
         dm : `holoviews.core.DynamicMap`
             The plot that automatically updates every `update_interval`.
         """
-        return live_plot(self, plotter=plotter,
-                         update_interval=update_interval,
-                         name=name)
+        return live_plot(
+            self, plotter=plotter, update_interval=update_interval, name=name
+        )
 
     def live_info(self, *, update_interval=0.1):
         """Display live information about the runner.
@@ -540,14 +625,14 @@ class AsyncRunner(BaseRunner):
         first_completed = asyncio.FIRST_COMPLETED
 
         if self._get_max_tasks() < 1:
-            raise RuntimeError('Executor has no workers')
+            raise RuntimeError("Executor has no workers")
 
         try:
             while not self.goal(self.learner):
                 futures = self._get_futures()
-                done, _ = await asyncio.wait(futures,
-                                             return_when=first_completed,
-                                             loop=self.ioloop)
+                done, _ = await asyncio.wait(
+                    futures, return_when=first_completed, loop=self.ioloop
+                )
                 self._process_futures(done)
         finally:
             remaining = self._remove_unfinished()
@@ -585,11 +670,13 @@ class AsyncRunner(BaseRunner):
         ...     save_kwargs=dict(fname='data/test.pickle'),
         ...     interval=600)
         """
+
         async def _saver(save_kwargs=save_kwargs, interval=interval):
-            while self.status() == 'running':
+            while self.status() == "running":
                 self.learner.save(**save_kwargs)
                 await asyncio.sleep(interval)
             self.learner.save(**save_kwargs)  # one last time
+
         self.saving_task = self.ioloop.create_task(_saver())
         return self.saving_task
 
@@ -640,6 +727,46 @@ def replay_log(learner, log):
         getattr(learner, method)(*args)
 
 
+# --- Useful runner goals
+
+
+def stop_after(*, seconds=0, minutes=0, hours=0):
+    """Stop a runner after a specified time.
+
+    For example, to specify a runner that should stop after
+    5 minutes, one could do the following:
+
+    >>> runner = Runner(learner, goal=stop_after(minutes=5))
+
+    To stop a runner after 2 hours, 10 minutes and 3 seconds,
+    one could do the following:
+
+    >>> runner = Runner(learner, goal=stop_after(hours=2, minutes=10, seconds=3))
+
+    Parameters
+    ----------
+    seconds, minutes, hours : float, default: 0
+        If more than one is specified, then they are added together
+
+    Returns
+    -------
+    goal : callable
+        Can be used as the ``goal`` parameter when constructing
+        a `Runner`.
+
+    Notes
+    -----
+    The duration specified is only a *lower bound* on the time that the
+    runner will run for, because the runner only checks its goal when
+    it adds points to its learner
+    """
+    stop_time = time.time() + seconds + 60 * minutes + 3600 * hours
+    return lambda _: time.time() > stop_time
+
+
+# -- Internal executor-related, things
+
+
 class SequentialExecutor(concurrent.Executor):
     """A trivial executor that runs functions synchronously.
 
@@ -663,7 +790,7 @@ class SequentialExecutor(concurrent.Executor):
 
 def _ensure_executor(executor):
     if executor is None:
-        executor = _default_executor(**_default_executor_kwargs)
+        executor = _default_executor()
 
     if isinstance(executor, concurrent.Executor):
         return executor
@@ -672,21 +799,28 @@ def _ensure_executor(executor):
     elif with_distributed and isinstance(executor, distributed.Client):
         return executor.get_executor()
     else:
-        raise TypeError('Only a concurrent.futures.Executor, distributed.Client,'
-                        ' or ipyparallel.Client can be used.')
+        raise TypeError(
+            "Only a concurrent.futures.Executor, distributed.Client,"
+            " or ipyparallel.Client can be used."
+        )
 
 
 def _get_ncores(ex):
     """Return the maximum  number of cores that an executor can use."""
     if with_ipyparallel and isinstance(ex, ipyparallel.client.view.ViewExecutor):
         return len(ex.view)
-    elif isinstance(ex, (concurrent.ProcessPoolExecutor,
-                         concurrent.ThreadPoolExecutor)):
+    elif isinstance(
+        ex, (concurrent.ProcessPoolExecutor, concurrent.ThreadPoolExecutor)
+    ):
+        return ex._max_workers  # not public API!
+    elif with_loky and isinstance(ex, loky.reusable_executor._ReusablePoolExecutor):
         return ex._max_workers  # not public API!
     elif isinstance(ex, SequentialExecutor):
         return 1
     elif with_distributed and isinstance(ex, distributed.cfexecutor.ClientExecutor):
         return sum(n for n in ex._client.ncores().values())
+    elif with_mpi4py and isinstance(ex, mpi4py.futures.MPIPoolExecutor):
+        ex.bootup()  # wait until all workers are up and running
+        return ex._pool.size  # not public API!
     else:
-        raise TypeError('Cannot get number of cores for {}'
-                        .format(ex.__class__))
+        raise TypeError(f"Cannot get number of cores for {ex.__class__}")
