@@ -1,10 +1,10 @@
-# -*- coding: utf-8 -*-
-
 import random
 
+import flaky
 import numpy as np
 
 from adaptive.learner import AverageLearner
+from adaptive.runner import simple
 
 
 def test_only_returns_new_points():
@@ -23,6 +23,7 @@ def test_only_returns_new_points():
     assert learner.ask(1)[0][0] == 10
 
 
+@flaky.flaky(max_runs=3)
 def test_avg_std_and_npoints():
     learner = AverageLearner(lambda x: x, atol=None, rtol=0.01)
 
@@ -42,7 +43,19 @@ def test_avg_std_and_npoints():
         if i > 2 and i % 10 == 0:
             # We need more than two points for 'learner.std' to be defined.
             values = np.array(list(learner.data.values()))
-            std = np.sqrt(sum((values - values.mean())**2) / (len(values) - 1))
+            std = np.sqrt(sum((values - values.mean()) ** 2) / (len(values) - 1))
             assert learner.npoints == len(learner.data)
             assert abs(learner.sum_f - values.sum()) < 1e-13
             assert abs(learner.std - std) < 1e-13
+
+
+def test_min_npoints():
+    def constant_function(seed):
+        return 0.1
+
+    for min_npoints in [1, 2, 3]:
+        learner = AverageLearner(
+            constant_function, atol=0.01, rtol=0.01, min_npoints=min_npoints
+        )
+        simple(learner, lambda l: l.loss() < 1)
+        assert learner.npoints >= max(2, min_npoints)
