@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 import collections
 import functools as ft
 import inspect
@@ -11,40 +9,57 @@ import random
 import shutil
 import tempfile
 
+import flaky
 import numpy as np
 import pytest
 import scipy.spatial
 
 import adaptive
-from adaptive.learner import (AverageLearner, BalancingLearner, DataSaver,
-    IntegratorLearner, Learner1D, Learner2D, LearnerND)
+from adaptive.learner import (
+    AverageLearner,
+    BalancingLearner,
+    DataSaver,
+    IntegratorLearner,
+    Learner1D,
+    Learner2D,
+    LearnerND,
+    SequenceLearner,
+)
 from adaptive.runner import simple
 
-
 try:
-    import skopt
-    from adaptive.learner import SKOptLearner
-except ModuleNotFoundError:
+    from adaptive.learner.skopt_learner import SKOptLearner
+except (ModuleNotFoundError, ImportError):
+    # XXX: catch the ImportError because of https://github.com/scikit-optimize/scikit-optimize/issues/902
     SKOptLearner = None
 
 
 LOSS_FUNCTIONS = {
-    Learner1D: ('loss_per_interval', (
-        adaptive.learner.learner1D.default_loss,
-        adaptive.learner.learner1D.uniform_loss,
-        adaptive.learner.learner1D.curvature_loss_function(),
-    )),
-    Learner2D: ('loss_per_triangle', (
-        adaptive.learner.learner2D.default_loss,
-        adaptive.learner.learner2D.uniform_loss,
-        adaptive.learner.learner2D.minimize_triangle_surface_loss,
-        adaptive.learner.learner2D.resolution_loss_function(),
-    )),
-    LearnerND: ('loss_per_simplex', (
-        adaptive.learner.learnerND.default_loss,
-        adaptive.learner.learnerND.std_loss,
-        adaptive.learner.learnerND.uniform_loss,
-    )),
+    Learner1D: (
+        "loss_per_interval",
+        (
+            adaptive.learner.learner1D.default_loss,
+            adaptive.learner.learner1D.uniform_loss,
+            adaptive.learner.learner1D.curvature_loss_function(),
+        ),
+    ),
+    Learner2D: (
+        "loss_per_triangle",
+        (
+            adaptive.learner.learner2D.default_loss,
+            adaptive.learner.learner2D.uniform_loss,
+            adaptive.learner.learner2D.minimize_triangle_surface_loss,
+            adaptive.learner.learner2D.resolution_loss_function(),
+        ),
+    ),
+    LearnerND: (
+        "loss_per_simplex",
+        (
+            adaptive.learner.learnerND.default_loss,
+            adaptive.learner.learnerND.std_loss,
+            adaptive.learner.learnerND.uniform_loss,
+        ),
+    ),
 }
 
 
@@ -60,8 +75,9 @@ def generate_random_parametrization(f):
     """
     _, *params = inspect.signature(f).parameters.items()
     if any(not callable(v.annotation) for (p, v) in params):
-        raise TypeError('All parameters to {} must be annotated with functions.'
-                        .format(f.__name__))
+        raise TypeError(
+            f"All parameters to {f.__name__} must be annotated with functions."
+        )
     realization = {p: v.annotation() for (p, v) in params}
     return ft.partial(f, **realization)
 
@@ -76,7 +92,6 @@ learner_function_combos = collections.defaultdict(list)
 
 
 def learn_with(learner_type, **init_kwargs):
-
     def _(f):
         learner_function_combos[learner_type].append((f, init_kwargs))
         return f
@@ -95,35 +110,40 @@ def maybe_skip(learner):
 # All parameters except the first must be annotated with a callable that
 # returns a random value for that parameter.
 
+
 @learn_with(Learner1D, bounds=(-1, 1))
-def quadratic(x, m: uniform(0, 10), b: uniform(0, 1)):
-    return m * x**2 + b
+def quadratic(x, m: uniform(1, 4), b: uniform(0, 1)):
+    return m * x ** 2 + b
 
 
 @learn_with(Learner1D, bounds=(-1, 1))
+@learn_with(SequenceLearner, sequence=np.linspace(-1, 1, 201))
 def linear_with_peak(x, d: uniform(-1, 1)):
     a = 0.01
-    return x + a**2 / (a**2 + (x - d)**2)
+    return x + a ** 2 / (a ** 2 + (x - d) ** 2)
 
 
 @learn_with(LearnerND, bounds=((-1, 1), (-1, 1)))
 @learn_with(Learner2D, bounds=((-1, 1), (-1, 1)))
+@learn_with(SequenceLearner, sequence=np.random.rand(1000, 2))
 def ring_of_fire(xy, d: uniform(0.2, 1)):
     a = 0.2
     x, y = xy
-    return x + math.exp(-(x**2 + y**2 - d**2)**2 / a**4)
+    return x + math.exp(-((x ** 2 + y ** 2 - d ** 2) ** 2) / a ** 4)
 
 
 @learn_with(LearnerND, bounds=((-1, 1), (-1, 1), (-1, 1)))
-def sphere_of_fire(xyz, d: uniform(0.2, 1)):
+@learn_with(SequenceLearner, sequence=np.random.rand(1000, 3))
+def sphere_of_fire(xyz, d: uniform(0.2, 0.5)):
     a = 0.2
     x, y, z = xyz
-    return x + math.exp(-(x**2 + y**2 + z**2 - d**2)**2 / a**4) + z**2
+    return x + math.exp(-((x ** 2 + y ** 2 + z ** 2 - d ** 2) ** 2) / a ** 4) + z ** 2
 
 
+@learn_with(SequenceLearner, sequence=range(1000))
 @learn_with(AverageLearner, rtol=1)
 def gaussian(n):
-    return random.gauss(0, 1)
+    return random.gauss(1, 1)
 
 
 # Decorators for tests.
@@ -141,21 +161,20 @@ def add_loss_to_params(learner_type, existing_params):
 
 def run_with(*learner_types, with_all_loss_functions=True):
     pars = []
-    for l in learner_types:
-        has_marker = isinstance(l, tuple)
+    for learner in learner_types:
+        has_marker = isinstance(learner, tuple)
         if has_marker:
-            marker, l = l
-        for f, k in learner_function_combos[l]:
-            ks = add_loss_to_params(l, k) if with_all_loss_functions else [k]
+            marker, learner = learner
+        for f, k in learner_function_combos[learner]:
+            ks = add_loss_to_params(learner, k) if with_all_loss_functions else [k]
             for k in ks:
                 # Check if learner was marked with our `xfail` decorator
                 # XXX: doesn't work when feeding kwargs to xfail.
                 if has_marker:
-                    pars.append(pytest.param(l, f, dict(k),
-                                             marks=[marker]))
+                    pars.append(pytest.param(learner, f, dict(k), marks=[marker]))
                 else:
-                    pars.append((l, f, dict(k)))
-    return pytest.mark.parametrize('learner_type, f, learner_kwargs', pars)
+                    pars.append((learner, f, dict(k)))
+    return pytest.mark.parametrize("learner_type, f, learner_kwargs", pars)
 
 
 def ask_randomly(learner, rounds, points):
@@ -163,17 +182,18 @@ def ask_randomly(learner, rounds, points):
     n_points = [random.randrange(*points) for _ in range(n_rounds)]
 
     xs = []
-    ls = []
+    losses = []
     for n in n_points:
-        x, l = learner.ask(n)
-        xs.extend(x)
-        ls.extend(l)
+        new_xs, new_losses = learner.ask(n)
+        xs.extend(new_xs)
+        losses.extend(new_losses)
 
-    return xs, ls
+    return xs, losses
 
 
 # Tests
 
+
 @run_with(Learner1D)
 def test_uniform_sampling1D(learner_type, f, learner_kwargs):
     """Points are sampled uniformly if no data is provided.
@@ -207,20 +227,23 @@ def test_uniform_sampling2D(learner_type, f, learner_kwargs):
 
     # regular grid
     n = math.sqrt(len(points))
-    xbounds, ybounds = learner_kwargs['bounds']
+    xbounds, ybounds = learner_kwargs["bounds"]
     r = math.sqrt((ybounds[1] - ybounds[0]) / (xbounds[1] - xbounds[0]))
     xs, dx = np.linspace(*xbounds, int(n / r), retstep=True)
     ys, dy = np.linspace(*ybounds, int(n * r), retstep=True)
 
     distances, neighbors = tree.query(list(it.product(xs, ys)), k=1)
-    assert max(distances) < math.sqrt(dx**2 + dy**2)
+    assert max(distances) < math.sqrt(dx ** 2 + dy ** 2)
 
 
-@pytest.mark.parametrize('learner_type, bounds', [
-    (Learner1D, (-1, 1)),
-    (Learner2D, [(-1, 1), (-1, 1)]),
-    (LearnerND, [(-1, 1), (-1, 1), (-1, 1)]),
-])
+@pytest.mark.parametrize(
+    "learner_type, bounds",
+    [
+        (Learner1D, (-1, 1)),
+        (Learner2D, [(-1, 1), (-1, 1)]),
+        (LearnerND, [(-1, 1), (-1, 1), (-1, 1)]),
+    ],
+)
 def test_learner_accepts_lists(learner_type, bounds):
     def f(x):
         return [0, 1]
@@ -229,7 +252,7 @@ def test_learner_accepts_lists(learner_type, bounds):
     simple(learner, goal=lambda l: l.npoints > 10)
 
 
-@run_with(Learner1D, Learner2D, LearnerND)
+@run_with(Learner1D, Learner2D, LearnerND, SequenceLearner)
 def test_adding_existing_data_is_idempotent(learner_type, f, learner_kwargs):
     """Adding already existing data is an idempotent operation.
 
@@ -246,7 +269,7 @@ def test_adding_existing_data_is_idempotent(learner_type, f, learner_kwargs):
     N = random.randint(10, 30)
     control.ask(N)
     xs, _ = learner.ask(N)
-    points = [(x, f(x)) for x in xs]
+    points = [(x, learner.function(x)) for x in xs]
 
     for p in points:
         control.tell(*p)
@@ -259,13 +282,24 @@ def test_adding_existing_data_is_idempotent(learner_type, f, learner_kwargs):
     M = random.randint(10, 30)
     pls = zip(*learner.ask(M))
     cpls = zip(*control.ask(M))
-    # Point ordering is not defined, so compare as sets
-    assert set(pls) == set(cpls)
+    if learner_type is SequenceLearner:
+        # The SequenceLearner's points might not be hasable
+        points, values = zip(*pls)
+        indices, points = zip(*points)
+
+        cpoints, cvalues = zip(*cpls)
+        cindices, cpoints = zip(*cpoints)
+        assert (np.array(points) == np.array(cpoints)).all()
+        assert values == cvalues
+        assert indices == cindices
+    else:
+        # Point ordering is not defined, so compare as sets
+        assert set(pls) == set(cpls)
 
 
-# XXX: This *should* pass (https://gitlab.kwant-project.org/qt/adaptive/issues/84)
+# XXX: This *should* pass (https://github.com/python-adaptive/adaptive/issues/55)
 #      but we xfail it now, as Learner2D will be deprecated anyway
-@run_with(Learner1D, xfail(Learner2D), LearnerND, AverageLearner)
+@run_with(Learner1D, xfail(Learner2D), LearnerND, AverageLearner, SequenceLearner)
 def test_adding_non_chosen_data(learner_type, f, learner_kwargs):
     """Adding data for a point that was not returned by 'ask'."""
     # XXX: learner, control and bounds are not defined
@@ -282,7 +316,7 @@ def test_adding_non_chosen_data(learner_type, f, learner_kwargs):
     N = random.randint(10, 30)
     xs, _ = control.ask(N)
 
-    ys = [f(x) for x in xs]
+    ys = [learner.function(x) for x in xs]
     for x, y in zip(xs, ys):
         control.tell(x, y)
         learner.tell(x, y)
@@ -290,9 +324,21 @@ def test_adding_non_chosen_data(learner_type, f, learner_kwargs):
     M = random.randint(10, 30)
     pls = zip(*learner.ask(M))
     cpls = zip(*control.ask(M))
-    # Point ordering within a single call to 'ask'
-    # is not guaranteed to be the same by the API.
-    assert set(pls) == set(cpls)
+
+    if learner_type is SequenceLearner:
+        # The SequenceLearner's points might not be hasable
+        points, values = zip(*pls)
+        indices, points = zip(*points)
+
+        cpoints, cvalues = zip(*cpls)
+        cindices, cpoints = zip(*cpoints)
+        assert (np.array(points) == np.array(cpoints)).all()
+        assert values == cvalues
+        assert indices == cindices
+    else:
+        # Point ordering within a single call to 'ask'
+        # is not guaranteed to be the same by the API.
+        assert set(pls) == set(cpls)
 
 
 @run_with(Learner1D, xfail(Learner2D), xfail(LearnerND), AverageLearner)
@@ -316,7 +362,7 @@ def test_point_adding_order_is_irrelevant(learner_type, f, learner_kwargs):
     N = random.randint(10, 30)
     control.ask(N)
     xs, _ = learner.ask(N)
-    points = [(x, f(x)) for x in xs]
+    points = [(x, learner.function(x)) for x in xs]
 
     for p in points:
         control.tell(*p)
@@ -336,9 +382,11 @@ def test_point_adding_order_is_irrelevant(learner_type, f, learner_kwargs):
 
 
 # XXX: the Learner2D fails with ~50% chance
-# see https://gitlab.kwant-project.org/qt/adaptive/issues/84
+# see https://github.com/python-adaptive/adaptive/issues/55
 @run_with(Learner1D, xfail(Learner2D), LearnerND, AverageLearner)
-def test_expected_loss_improvement_is_less_than_total_loss(learner_type, f, learner_kwargs):
+def test_expected_loss_improvement_is_less_than_total_loss(
+    learner_type, f, learner_kwargs
+):
     """The estimated loss improvement can never be greater than the total loss."""
     f = generate_random_parametrization(f)
     learner = learner_type(f, **learner_kwargs)
@@ -346,24 +394,27 @@ def test_expected_loss_improvement_is_less_than_total_loss(learner_type, f, lear
     xs, loss_improvements = learner.ask(N)
 
     for x in xs:
-        learner.tell(x, f(x))
+        learner.tell(x, learner.function(x))
 
     M = random.randint(50, 100)
     _, loss_improvements = learner.ask(M)
 
     if learner_type is Learner2D:
-        assert (sum(loss_improvements)
-                < sum(learner.loss_per_triangle(learner.ip())))
+        assert sum(loss_improvements) < sum(
+            learner.loss_per_triangle(learner.interpolator(scaled=True))
+        )
     elif learner_type is Learner1D:
         assert sum(loss_improvements) < sum(learner.losses.values())
     elif learner_type is AverageLearner:
         assert sum(loss_improvements) < learner.loss()
 
 
-# XXX: This *should* pass (https://gitlab.kwant-project.org/qt/adaptive/issues/84)
+# XXX: This *should* pass (https://github.com/python-adaptive/adaptive/issues/55)
 #      but we xfail it now, as Learner2D will be deprecated anyway
 @run_with(Learner1D, xfail(Learner2D), LearnerND)
-def test_learner_performance_is_invariant_under_scaling(learner_type, f, learner_kwargs):
+def test_learner_performance_is_invariant_under_scaling(
+    learner_type, f, learner_kwargs
+):
     """Learners behave identically under transformations that leave
        the loss invariant.
 
@@ -380,9 +431,8 @@ def test_learner_performance_is_invariant_under_scaling(learner_type, f, learner
     yscale = 1000 * random.random()
 
     l_kwargs = dict(learner_kwargs)
-    l_kwargs['bounds'] = xscale * np.array(l_kwargs['bounds'])
-    learner = learner_type(lambda x: yscale * f(np.array(x) / xscale),
-                           **l_kwargs)
+    l_kwargs["bounds"] = xscale * np.array(l_kwargs["bounds"])
+    learner = learner_type(lambda x: yscale * f(np.array(x) / xscale), **l_kwargs)
 
     if learner_type in [Learner1D, LearnerND]:
         learner._recompute_losses_factor = 1
@@ -408,12 +458,21 @@ def test_learner_performance_is_invariant_under_scaling(learner_type, f, learner
     assert math.isclose(learner.loss(), control.loss(), rel_tol=1e-10)
 
 
-@run_with(Learner1D, Learner2D, LearnerND, AverageLearner,
-    with_all_loss_functions=False)
+@flaky.flaky(max_runs=3)
+@run_with(
+    Learner1D,
+    Learner2D,
+    LearnerND,
+    AverageLearner,
+    SequenceLearner,
+    with_all_loss_functions=False,
+)
 def test_balancing_learner(learner_type, f, learner_kwargs):
     """Test if the BalancingLearner works with the different types of learners."""
-    learners = [learner_type(generate_random_parametrization(f), **learner_kwargs)
-                for i in range(4)]
+    learners = [
+        learner_type(generate_random_parametrization(f), **learner_kwargs)
+        for i in range(4)
+    ]
 
     learner = BalancingLearner(learners)
 
@@ -439,12 +498,21 @@ def test_balancing_learner(learner_type, f, learner_kwargs):
             x = stash.pop()
             learner.tell(x, learner.function(x))
 
-    assert all(l.npoints > 10 for l in learner.learners), [l.npoints for l in learner.learners]
+    assert all(l.npoints > 5 for l in learner.learners), [
+        l.npoints for l in learner.learners
+    ]
 
 
-@run_with(Learner1D, Learner2D, LearnerND, AverageLearner,
-    maybe_skip(SKOptLearner), IntegratorLearner,
-    with_all_loss_functions=False)
+@run_with(
+    Learner1D,
+    Learner2D,
+    LearnerND,
+    AverageLearner,
+    maybe_skip(SKOptLearner),
+    IntegratorLearner,
+    SequenceLearner,
+    with_all_loss_functions=False,
+)
 def test_saving(learner_type, f, learner_kwargs):
     f = generate_random_parametrization(f)
     learner = learner_type(f, **learner_kwargs)
@@ -454,6 +522,7 @@ def test_saving(learner_type, f, learner_kwargs):
         control._recompute_losses_factor = 1
     simple(learner, lambda l: l.npoints > 100)
     fd, path = tempfile.mkstemp()
+    os.close(fd)
     try:
         learner.save(path)
         control.load(path)
@@ -466,9 +535,16 @@ def test_saving(learner_type, f, learner_kwargs):
         os.remove(path)
 
 
-@run_with(Learner1D, Learner2D, LearnerND, AverageLearner,
-    maybe_skip(SKOptLearner), IntegratorLearner,
-    with_all_loss_functions=False)
+@run_with(
+    Learner1D,
+    Learner2D,
+    LearnerND,
+    AverageLearner,
+    maybe_skip(SKOptLearner),
+    IntegratorLearner,
+    SequenceLearner,
+    with_all_loss_functions=False,
+)
 def test_saving_of_balancing_learner(learner_type, f, learner_kwargs):
     f = generate_random_parametrization(f)
     learner = BalancingLearner([learner_type(f, **learner_kwargs)])
@@ -483,7 +559,7 @@ def test_saving_of_balancing_learner(learner_type, f, learner_kwargs):
     folder = tempfile.mkdtemp()
 
     def fname(learner):
-        return folder + 'test'
+        return folder + "test"
 
     try:
         learner.save(fname=fname)
@@ -497,13 +573,19 @@ def test_saving_of_balancing_learner(learner_type, f, learner_kwargs):
         shutil.rmtree(folder)
 
 
-@run_with(Learner1D, Learner2D, LearnerND, AverageLearner,
-    maybe_skip(SKOptLearner), IntegratorLearner,
-    with_all_loss_functions=False)
+@run_with(
+    Learner1D,
+    Learner2D,
+    LearnerND,
+    AverageLearner,
+    maybe_skip(SKOptLearner),
+    IntegratorLearner,
+    with_all_loss_functions=False,
+)
 def test_saving_with_datasaver(learner_type, f, learner_kwargs):
     f = generate_random_parametrization(f)
-    g = lambda x: {'y': f(x), 't': random.random()}
-    arg_picker = operator.itemgetter('y')
+    g = lambda x: {"y": f(x), "t": random.random()}  # noqa: E731
+    arg_picker = operator.itemgetter("y")
     learner = DataSaver(learner_type(g, **learner_kwargs), arg_picker)
     control = DataSaver(learner_type(g, **learner_kwargs), arg_picker)
 
@@ -513,6 +595,7 @@ def test_saving_with_datasaver(learner_type, f, learner_kwargs):
 
     simple(learner, lambda l: l.npoints > 100)
     fd, path = tempfile.mkstemp()
+    os.close(fd)
     try:
         learner.save(path)
         control.load(path)
